ax25 = {}

local function compose_ax25_address(address_object, last_address)
  local callsign = address_object.callsign
  local c_bit = address_object.c_bit or false
  local r_bits = address_object.r_bits or 0
  local ssid = address_object.ssid or 0
  local address = ""
  for i = 1, 6 do
    local char = string.sub(callsign, i, i)
    if char == "" then
      char = " "
    end
    address = address .. string.char(string.byte(char) << 1)
  end
  ssid = ssid << 1
  ssid = ssid | (c_bit and 0x01 or 0x00) << 7
  ssid = ssid | r_bits << 5
  if address_object.last_address or last_address then
    ssid = ssid | 0x01
  end
  address = address .. string.char(ssid)
  return address
end

local function parse_ax25_address(raw_address)
  if #raw_address ~= 7 then return nil end
  -- extract callsign
  local callsign = ""
  local space_occured = false
  for i = 1, 6 do
    local char = string.char(string.byte(string.sub(raw_address, i, i)) >> 1)
    if char ~= " " then
      if space_occured then
        return nil
      end
      callsign = callsign .. char
    else
      space_occured = true
    end
  end
  if not string.match(callsign, "^[0-9A-Z]+$") then
    return nil
  end
  -- extract ssid and flags
  local ssid = string.byte(string.sub(raw_address, 7, 7))
  local address = {
    callsign = callsign,
    c_bit = (ssid >> 7) & 0x01 == 0x01,
    r_bits = (ssid >> 5) & 0x03,
    ssid = (ssid >> 1) & 0x0f
  }
  address.station = address.callsign
  if address.ssid > 0 then
    address.station = address.station .. "-" .. address.ssid
  end
  -- extract last address flag
  local last_address = string.byte(string.sub(raw_address, 7, 7)) & 0x01 == 0x01
  address.last_address = last_address
  return address, last_address
end

function ax25.compose_frame(frame)
  local raw = ""
  raw = raw .. compose_ax25_address(frame.destination)
  raw = raw .. compose_ax25_address(frame.source)
  for i = 1, #frame.path do
    raw = raw .. compose_ax25_address(frame.path[i], i == #frame.path)
  end
  raw = raw .. frame.control
  raw = raw .. frame.pid
  raw = raw .. frame.info
  return raw
end

function ax25.parse_frame(raw)
  local frame = {}
  frame.raw = raw
  -- extract destination address
  frame.destination = parse_ax25_address(string.sub(raw, 1, 7))
  if frame.destination == nil then return end
  -- extract source address
  frame.source = parse_ax25_address(string.sub(raw, 8, 14))
  if frame.source == nil then return end
  frame.path_printable = frame.source.callsign
  if frame.source.ssid > 0 then
    frame.path_printable = frame.path_printable .. "-" .. frame.source.ssid
  end
  frame.path_printable = frame.path_printable .. ">" .. frame.destination.callsign 
  if frame.destination.ssid > 0 then
    frame.path_printable = frame.path_printable .. "-" .. frame.destination.ssid
  end
  -- extract path addresses
  frame.path = {}
  local pos = 15
  repeat
    local address, last_address = parse_ax25_address(string.sub(raw, pos, pos + 6))
    if address == nil then return end
    table.insert(frame.path, address)
    frame.path_printable = frame.path_printable .. "," .. address.callsign
    if address.ssid > 0 then
      frame.path_printable = frame.path_printable .. "-" .. address.ssid
    end
    pos = pos + 7
  until last_address
  frame.raw_printable = frame.path_printable
  -- extract control and pid
  frame.control = string.sub(raw, pos, pos)
  pos = pos + 1
  frame.pid = string.sub(raw, pos, pos)
  pos = pos + 1
  -- extract information block
  frame.info = string.sub(raw, pos, #raw)
  frame.raw_printable = frame.raw_printable .. ":" .. frame.info
  frame.raw_printable = string.gsub(frame.raw_printable, "<([0-9a-fA-F])([0-9a-fA-F])>", function (match1, match2)
    return string.format("<3c><%02x><%02x><3e>", string.byte(match1), string.byte(match2))
  end)  
  frame.raw_printable = string.gsub(frame.raw_printable, "[\000-\031\128-\255]", function (match)
    return string.format("<%02x>", string.byte(match))
  end)  
  return frame
end

function ax25.debug(frame)
  -- prints an ax25 frame
  from = frame.source.callsign
  if frame.source.ssid > 0 then
    from = from .. "-" .. frame.source.ssid
  end
  to = frame.destination.callsign
  if frame.destination.ssid > 0 then
    to = to .. "-" .. frame.destination.ssid
  end
  io.write(from, ">", to)
  for i, path in ipairs(frame.path) do
    local station = path.callsign
    if path.ssid then
      station = station .. "-" .. path.ssid
    end
    io.write(",", station)
  end
  local info = string.gsub(frame.info, "<([0-9a-fA-F])([0-9a-fA-F])>", function (match1, match2)
    return string.format("<3c><%02x><%02x><3e>", string.byte(match1), string.byte(match2))
  end)  
  info = string.gsub(info, "[\000-\031\128-\255]", function (match)
    return string.format("<%02x>", string.byte(match))
  end)  
  io.write(":", info, "\n")
end


