#!/usr/bin/env lua

-- SERVER CONFIG
-- --------------------------------------------------------------------------

-- routing path
APRS_PATH = { { callsign = "WIDE1", ssid = 1 }, { callsign = "WIDE2", ssid = 1 } }

-- during which interval two identical frames are treated as duplicates
DUPLICATE_FRAME_TIMEOUT = 180 -- seconds

-- callsign pattern, set this for non ham radio use, e.g. citizen band (CB)
CALLSIGN_DETECT_REGEX = "^[A-Z0-9][A-Z0-9][0-9][A-Z][A-Z]?[A-Z]?-?[0-9]?[0-9]?$" -- default
-- CALLSIGN_DETECT_REGEX = "^[0-9][0-9][A-Z][A-Z0-9]$"

-- software identifier (experimental)
APRS_SOFTWARE_IDENT = { callsign = "APZRID", ssid = 0 }

-- simulate packet loss (0 = no loss, 1 = total loss)
--SIMULATE_PACKET_LOSS_RATE = 0.5

-- send out how many repeats of a direct message if not acknowledged
DIRECT_MESSAGE_RETRY_COUNT = 7

-- how long to wait for a direct message ack
local function wait_for_ack_time(attempt)
  return math.floor(60 - math.log(1 - math.random()) * (56.25 * 2 ^ attempt - 60))
end

-- send out how many direct message acks if no subsequent message received
DIRECT_MESSAGE_ACK_RETRY_COUNT = 7

-- how long to wait before sending subsequent direct message ack
local function wait_next_ack_time(attempt)
  return math.floor(60 - math.log(1 - math.random()) * (56.25 * 2 ^ attempt - 60))
end

CHAT_RESEND_COUNT = 7

-- how long before send chat again
local function wait_next_chat_repeat_time(attempt)
  return math.floor(60 - math.log(1 - math.random()) * (56.25 * 2 ^ attempt - 60))
end

-- IRC CONFIG
-- --------------------------------------------------------------------------
server_version = "radioircd v0.3"
local hostname_fh = assert(io.popen("/bin/hostname"), "Could not execute /bin/hostname")
server_name = assert(hostname_fh:read("*l"), "Could not read output of /bin/hostname")
hostname_fh:close()
passwords = { }
max_line_length_in = 1024
max_line_length_out = 512
max_channel_count = 32
max_nick_length = 16
max_channel_length = 32
max_user_length = 16
connect_timeout = 60
ping_interval = 240
ping_timeout = 60
quit_timeout = 60


-- --------------------------------------------------------------------------
-- END OF CONFIG


-- set lua package path
package.path = "../lib/?.lua;" .. package.path

-- load libraries
require "kiss"
require "ax25"
require "aprs"

-- read command line arguments
local kiss_host, kiss_port, irc_port = ...
kiss_host = kiss_host or "localhost"
kiss_port = kiss_port or 8001
irc_port = irc_port or 6667


-- initialize random number generator
math.randomseed(os.time())

-- install TCP listeners
listeners = {
  moonbridge_io.tcplisten("::", irc_port),
  moonbridge_io.tcplisten("0.0.0.0", irc_port)
}

-- load TLS certificates
--local tlsconfig = assert(moonbridge_io.tlsconf{
--  mode = "server",
--  cert_file = "selfsigned.pem",
--  key_file = "selfsigned.pem"
--})

-- set of symbols for message IDs
local compressed_id_chars = 
  "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

local frames_heared = {}

local stations_heared = {}

-- queues of outgoing direct messages
local message_out_queues = {}

-- queues of outgoing direct message acks
local message_ack_out_reset_functions = {}


local chat_out_ids = {}

local chat_in_last_ids = {}
local chat_out_last_messages = {}

local chat_out_reset_functions = {}


local function escape_chat_message(message)
  return (string.gsub(message, "\\", "\\\\"):gsub("|", "\\p"))
end

local function unescape_chat_message(message)
  return (string.gsub(message, "\\p", "|"):gsub("\\\\", "\\"))
end

local function check_channelname(name)
  if string.match(upper(name), "^[0-9A-Z_.-]+$") then
    return true
  end
end

local function extract_callsign(station)
  local callsign, ssid = string.match(
    station, "^([0-9A-Z][0-9A-Z]?[A-Z0-9]?[0-9A-Z]?[0-9A-Z]?[0-9A-Z]?)-([0-9][0-9]?)$"
  )
  if not ssid then
    callsign = string.match(
      station, "^([0-9A-Z][0-9A-Z]?[A-Z0-9]?[0-9A-Z]?[0-9A-Z]?[0-9A-Z]?)$"
    )
  end
  if ssid == nil then
    ssid = 0
  else
    ssid = tonumber(ssid)
  end
  return callsign, ssid
end

local function normalize_station(station)
  local callsign, ssid = extract_callsign(station)
  if ssid == 0 then
    return callsign
  end
  return station
end

-- prepare AX.25 frame
local function prepare_ax25_frame(from)
  local callsign, ssid = extract_callsign(from)
  return {
    source = { callsign = callsign, ssid = ssid },
    destination = APRS_SOFTWARE_IDENT,
    path = APRS_PATH,
    control = string.char(0x03),
    pid = string.char(0xf0), 
  }
end

-- send a message to an IRC channel
local function send_to_channel(prefix, channel_name, from, message)
  local channel_name = lower(channel_name)
  local channel = channels[prefix .. channel_name]
  if not channel then return end
  channel:send(
    ":", from .. "!extern@local PRIVMSG " .. prefix .. upper(channel_name) .. 
    " :", message
  )
end

local function decompress_id(compressed_id)
  local id = 0
  for i = #compressed_id, 1, -1 do
    local symbol = string.sub(compressed_id, i, i)
    local escaped_symbol = string.gsub(symbol, "[^0-9A-Za-z]", "%%%1")
    local symbol_id = string.find(compressed_id_chars, escaped_symbol)
    if not symbol_id then return nil end
    id = id + symbol_id * (#compressed_id_chars) ^ (#compressed_id - i)
  end
  return id
end

-- encode numerical message ID
local function compress_id(length, id)
  local compressed_id = ""
  while length > 0 do
    local symbol_id = id % #compressed_id_chars
    compressed_id = string.sub(compressed_id_chars, symbol_id+1, symbol_id+1)
                    .. compressed_id
    id = id // #compressed_id_chars
    length = length - 1
  end
  return compressed_id
end

-- increment numerical message ID by one with wraparound
local function increment_id(length, id)
  return (id + 1) % (#compressed_id_chars) ^ length
end

local function id_difference(length, id1, id2)
  local id1 = decompress_id(id1)
  local id2 = decompress_id(id2)
  if not id1 or not id2 then
    return
  end
  local difference = (id2 - id1) % (#compressed_id_chars) ^ length
  return difference
end


-- create a random message ID
local function random_id(length)
  return math.random(0, (#compressed_id_chars) ^ length - 1)
end

-- returns the next message ID for a queue
local function next_message_id(queue, message)
  if queue.id then
    queue.id = increment_id(2, queue.id)
  else
    queue.id = random_id(2)
  end
  return compress_id(2, queue.id)
end

-- returns the next chat ID for a channel
local function next_chat_id(from, channel_name)
  local from = normalize_station(from)
  local id = chat_out_ids[from .. " " .. channel_name]
  if id then
    id = increment_id(4, id)
  else
    id = random_id(4)
  end
  chat_out_ids[from .. " " .. channel_name] = id
  return compress_id(4, id)
end

local function send_kiss_frame(data)
  if SIMULATE_PACKET_LOSS_RATE then
    if math.random(1, 1/SIMULATE_PACKET_LOSS_RATE) == 1 then
      return
    end
  end
  uplink:flush(data)
end

-- send a direct message out as AX.25 frame via KISS    
local function send_message(from, to, message, id)
  local ax25_frame = prepare_ax25_frame(from)
  ax25_frame.info = aprs.compose_message(to, message, id)
  local ax25_data = ax25.compose_frame(ax25_frame)
  local kiss_data = kiss.compose_frame(ax25_data)
  io.write("tx ")
  ax25.debug(ax25_frame)
  local frame = ax25.parse_frame(ax25_data)
  table.insert(frames_heared, {
    frame = frame.source.station .. " " .. frame.destination.station .. " " .. frame.info,
    received = os.time()
  })
  send_kiss_frame(kiss_data)
end

-- drop a direct message queue
local function drop_queue(client, from, to)
  if client then
    client:send_notice(
      from, " ", to, 
      " message queue dropped, no acks received..."
    )
  end
  local queue = message_out_queues[from .. " " .. to]
  queue.timeout_reset()
  queue.messages = {}  
end

-- handle entries of a direct message queue
local function process_queue(client, from, to, attempt)
  local from = normalize_station(from)
  local to = normalize_station(to)
  local attempt = attempt or 1
  local queue = message_out_queues[from .. " " .. to]
  if not queue.messages[1] then return end
  send_message(from, to, queue.messages[1].message, queue.messages[1].id)
  local wait_for_ack_time = wait_for_ack_time(attempt)
  --[[
  if client then
    client:send_notice(
      "TX ", from, " ", to, " ", 
      queue.messages[1].message, 
      " (ID: ", queue.messages[1].id, ", ", attempt, ". attempt, waiting ",
      wait_for_ack_time, " seconds for ack)"
    )
  end
  --]]
  if attempt < DIRECT_MESSAGE_RETRY_COUNT then
    queue.timeout_reset = timeout(wait_for_ack_time, function()
      process_queue(client, from, to, attempt + 1)
    end)
  else
    queue.timeout_reset = timeout(wait_for_ack_time, function()
      drop_queue(client, from, to)
    end)
  end
end

-- add an aprs direct message to the queue
local function queue_message(client, from, to, message)
  local from = normalize_station(from)
  local to = normalize_station(to)
  local queue = message_out_queues[from .. " " .. to] or { 
    messages = {},
    timeout_reset = function() end
  }
  local id = next_message_id(queue)
  table.insert(queue.messages, { message = message, id = id })
  message_out_queues[from .. " " .. to] = queue
  if #queue.messages == 1 then
    process_queue(client, from, to)
  end
end

-- handle incoming aprs ack messages
local function process_message_ack(from, to, id)
  local from = normalize_station(from)
  local to = normalize_station(to)
  local queue = message_out_queues[to .. " " .. from]
  if not queue or not queue.messages[1] then return end
  if queue.messages[1].id == id then
    table.remove(queue.messages, 1)
    queue.timeout_reset()
    process_queue(nil, to, from)
  end
end

-- handle direct message ack sending
local function ack_message(client, from, to, id, attempt)
  local from = normalize_station(from)
  local to = normalize_station(to)
  local attempt = attempt or 1
  send_message(from, to, "ack" .. id)
  local wait_for_next_ack_time = wait_next_ack_time(attempt)
  --[[
  if client then
    client:send_notice(
      "TX ", from, " ", to, " ", 
      "acknowledge message ID ", id, " (", attempt, ". attempt)"
    )
  end
  --]]
  if message_ack_out_reset_functions[from .. " " .. to] then
    message_ack_out_reset_functions[from .. " " .. to]()
  end
  if attempt < DIRECT_MESSAGE_ACK_RETRY_COUNT then
    message_ack_out_reset_functions[from .. " " .. to] = timeout(wait_for_next_ack_time, function()
      ack_message(client, from, to, id, attempt + 1)
    end)
  end
end

local function process_send_chat(from, channel_name, id, attempt)
  local attempt = attempt or 1
  local last_messages = chat_out_last_messages[from .. " " .. channel_name]
  local message = ""
  for i, last_message in ipairs(last_messages) do
    if last_message ~= "" then
      message = message .. "|" .. escape_chat_message(last_message)
    end
  end
  local max_message_length = 255 - 5 - #channel_name
  if #message > max_message_length then
    message = string.sub(message, #message - max_message_length + 1, #message)
  end
  message = string.sub(message, string.find(message, "|"), #message)
  local ax25_frame = prepare_ax25_frame(from)
  ax25_frame.info = aprs.compose_chat(channel_name, message, id)
  local ax25_data = ax25.compose_frame(ax25_frame)
  local kiss_data = kiss.compose_frame(ax25_data)
  io.write("tx ")
  ax25.debug(ax25_frame)
  local frame = ax25.parse_frame(ax25_data)
  table.insert(frames_heared, {
    frame = frame.source.station .. " " .. frame.destination.station .. " " .. frame.info,
    received = os.time()
  })
  send_kiss_frame(kiss_data)
  if attempt < CHAT_RESEND_COUNT then
    chat_out_reset_functions[from .. " " .. channel_name] = timeout(wait_next_chat_repeat_time(attempt), function()
      local last_messages = chat_out_last_messages[from .. " " .. channel_name]
      if #last_messages > 1 then
        table.remove(last_messages, 1)
        chat_out_last_messages[from .. " " .. channel_name] = last_messages
        process_send_chat(from, channel_name, id, attempt + 1)
      end
    end)
  end
end

local function send_chat(from, channel_name, message)
  local id = next_chat_id(from, channel_name)
  if not check_channelname(channel_name) then
    return false
  end
  if #message > 255 - 5 - #channel_name then
    message = string.sub(message, #message > 255 - 5 - #channel_name, 255)
  end
  local last_messages = chat_out_last_messages[from .. " " .. channel_name] or { }
  for i = 1, CHAT_RESEND_COUNT - #last_messages  do
    table.insert(last_messages, "")
  end
  table.insert(last_messages, message)
  table.remove(last_messages, 1)
  chat_out_last_messages[from .. " " .. channel_name] = last_messages
  if chat_out_reset_functions[from .. " " .. channel_name] then
    chat_out_reset_functions[from .. " " .. channel_name]()
  end
  process_send_chat(from, channel_name, id)
end

-- only even frames are AX.25 content
local even = false

-- define handler for incoming AX.25 frames via KISS
uplink = {
  socket = assert(moonbridge_io.tcpconnect(kiss_host, kiss_port)),
  terminator = string.char(0xc0),
  handler = function(kiss_frame)
    even = not even
    if even then return end

    -- add leading FEND
    kiss_frame = kiss.FEND .. kiss_frame

    -- decode KISS frame to AX.25 frame
    local port, command, ax25_data = kiss.parse_frame(kiss_frame)
    if ax25_data == nil then return end

    -- decode AX.25 frame
    local frame = ax25.parse_frame(ax25_data)
    if not frame then return end

    -- output received frame
    io.write("rx ")
    ax25.debug(frame)

    -- parse aprs body
    local message = aprs.parse(frame.info)
  
    -- update heared stations table
    local heared_station = stations_heared[upper(frame.source.station)] or {
      via_paths = {},
      to_channels = {},
      to_stations = {}
    }
    heared_station.via_paths[frame.path_printable] = os.time()
    if message then
      if message.message_type == "chat" then
        heared_station.to_channels[message.channel_name] = os.time()
      elseif message.message_type == "message" then
        heared_station.to_stations[message.station] = os.time()
      end
    end
    stations_heared[upper(frame.source.station)] = heared_station

    -- send all AX.25 frames to &raw as printable raw
    send_to_channel("&", "raw", frame.source.station, frame.raw_printable)

    -- check for duplicates
    local current_frame = {
      frame = frame.source.station .. " " .. frame.destination.station .. " " .. frame.info,
      received = os.time()
    }
    for i, check_frame in ipairs(frames_heared) do
      if os.time() - check_frame.received > DUPLICATE_FRAME_TIMEOUT then
        table.remove(frames_heared, i)
      else
        if check_frame.frame == current_frame.frame then -- duplicate detected
          -- send all AX.25 dupe frames to &raw-duplicates as printable raw
          send_to_channel("&", "raw-duplicates", frame.source.station, frame.raw_printable)
          return
        end
      end
    end
    table.insert(frames_heared, current_frame)

    -- send all AX.25 unique frames to &raw-unique as printable raw
    send_to_channel("&", "raw-unique", frame.source.station, frame.raw_printable)


    -- stop processing if no APRS message
    if not message then return end

    -- process direct messages 
    if message.message_type == "message" then

      -- check for direct message acks
      local ack_id = string.match(message.message, "^ack(..)$")
      if ack_id then
        process_message_ack(frame.source.station, message.station, ack_id)
      end

      if string.match(message.station, CALLSIGN_DETECT_REGEX) then
        -- send all direct messages to channel &messages
        send_to_channel(
          "&", "messages", frame.source.station, message.station .. ": " ..
          message.message
        )
      else
        -- send all bulletins to channel &bulletins
        send_to_channel(
          "&", "bulletins", frame.source.station, 
          message.station .. ": " .. message.message
        )
      end

      -- send all direct messages to + channel based on destination
      send_to_channel(
        "+", message.station, frame.source.station, message.message
      )
  
      -- send direct message to corresponding client, if available
      local client, client0
      client = nicks[string.lower(message.callsign) .. "-" .. message.ssid]
      if message.ssid == 0 then
        client0 = nicks[string.lower(message.callsign)]
      end
      if client or client0 then
        local function send(client)
          if ack_id then
            --client:send_notice("rx ", frame.source.station, " ", message.station, " ack for message ID ", ack_id, " received")
          else
            client:send(
              ":", frame.source.station .. "!" .. frame.source.station .. 
              "@" .. frame.path_printable .. 
              " PRIVMSG " .. message.station .. ": " .. message.message
            )
          end
        end
        if client then
          send(client)
          if message.id and not ack_id then
            ack_message(client, message.station, frame.source.station, message.id)
          end
        end
        if client0 then
          send(client0)
          if message.id and not ack_id then
            ack_message(client0, message.station, frame.source.station, message.id)
          end
        end
      end

    -- process chats
    elseif message.message_type == "chat" then

      local difference
      local last_id = chat_in_last_ids[message.channel_name .. " " .. frame.source.station]
      if last_id then
        difference = id_difference(4, last_id, message.id) or difference
      end
      local out_lines = {}
      if difference == 1 then
        local current_line = string.match(message.message, "|([^|]+)$")
        if current_line then
          table.insert(out_lines, unescape_chat_message(current_line))
        end
      else
        local lines = {}
        for line in string.gmatch(message.message, "|([^|]+)") do
          table.insert(lines, unescape_chat_message(line))
        end
        local start = 1
        if difference then
          start = #lines - difference + 1
        end
        for i = start, #lines do
          table.insert(out_lines, lines[i])
        end
      end
      chat_in_last_ids[message.channel_name .. " " .. frame.source.station] = message.id

      for i, line in ipairs(out_lines) do
        -- send all chats to channel &chats
        send_to_channel("&", "chats", frame.source.station, "#" .. message.channel_name ..": " .. line)

        -- send chats to corresponding channel
        send_to_channel("#", message.channel_name, frame.source.station, line)
      end
    end

  end
}

-- warn if nick is not a valid callsign
function client_prototype:nick_hook(nick)
  if not self.nick then
    return
  end
  if not string.match(upper(nick), CALLSIGN_DETECT_REGEX) then
    self:send_notice(
      "Nick is not a valid callsign / station identifier. NOT on air!"
    )
  else
    local on_air = [[
       _   _     _   _   _  
      / \ / \   / \ / \ / \ 
     ( O | N ) ( A | I | R )
      \_/ \_/   \_/ \_/ \_/ 
]]

    for line in string.gmatch(on_air, "[^\n]*") do
      self:send_notice(line)
    end
  end
end

function client_prototype:send_motd()
  local motd = [[
                   ___       _               __
   _________ _____/ (_)___  (_)_____________/ /
  / ___/ __ `/ __  / / __ \/ / ___/ ___/ __  / 
 / /  / /_/ / /_/ / / /_/ / / /  / /__/ /_/ /  
/_/   \__,_/\__,_/_/\____/_/_/   \___/\__,_/ v0.3
        
> Group chat with other stations:
  /join #<channelname> e.g. /join #test

> Direct chat with a station:
  /query <CALLSIGN>-<SSID> e.g. /query DL1PZ-7

> Read and send bulletins and announcements
  /join +<BULLETIN_ID> e.g. /join +BLN0TEST 

> Special channels
  /join &chats to read all group chats
  /join &messages to read all direct messages
  /join &bulletins to read all bulletins and announcements
  /join &raw to read all received frames including duplicates
  /join &raw-unique to read all received frames without duplicates
  /join &raw-duplicates to read only duplicates
  /join &raw-tx to send raw frames

Thank you for using radioircd! vy 73 de DL1PZ

]]
  self:send(375, ":- ", server_name, " Message of the day - ")
  for line in string.gmatch(motd, "([^\n]*)\n") do
    self:send(372, ":- ", line)
  end
  self:send(376, ":End of MOTD command")
end

function client_prototype:registered_hook()
  self:send_motd()
  self:nick_hook(self.nick)
  return true
end

function client_prototype:join_hook(channel, channel_name)
  if #channel.clients == 0 then
    local topics = {
      ["&raw"] = "All received AX.25 frames",
      ["&raw-unique"] = "All received AX.25 frames without duplicates",
      ["&raw-duplicates"] = "Only duplicates of received AX.25 frames",
      ["&raw-tx"] = "Send raw AX.25 frames (enter info content) [ON AIR]",
      ["&chats"] = "All received chat messages",
      ["&messages"] = "All received direct messages",
      ["&bulletins"] = "All received bulletins and announcements"
    }
    local topic = topics[channel.lower_name]
    if not topic then
      local prefix, target = string.match(channel.lower_name, "^(.)(.+)$")
      if prefix == "#" then
        topic = "Group chat #" .. target .. " [ON AIR]"
      elseif prefix == "+" then
        topic = "Bulletin / announcement " .. upper(target) .. " [ON AIR]"
      end
    end
    if topic then
      channel.topic = topic
    end
  end
end

function client_prototype:command_hook()

  if self:is_command("WHOIS", 1) then

    local targets = self.params[2] or self.params[1]
    local any_match = false
    local no_match = {}
    for target in targets:gmatch("[^,]+") do
      local target = upper(target)
      local match = false
      local heared_station = stations_heared[target]
      if heared_station then
        any_match = true
        match = true
        self:send(311, target, " ", target, " ", "air", " * :", target)
        local idletime = math.maxinteger
        for via_path, last_heared in pairs(heared_station.via_paths) do
          heared_ago = os.time() - last_heared
          idletime = math.min(idletime, heared_ago)
          local line = via_path .. " (" .. heared_ago .. " seconds ago)"
          self:send(312, target, " AIR :", line)
        end
        for channel, last_heared in pairs(heared_station.to_channels) do
          heared_ago = os.time() - last_heared
          local line = "#" .. channel .. " (" .. heared_ago .. " seconds ago)"
          self:send(319, target, " :", line )
        end 
        for station, last_heared in pairs(heared_station.to_stations) do
          heared_ago = os.time() - last_heared
          local line = "+" .. station .. " (" .. heared_ago .. " seconds ago)"
          self:send(319, target, " :", line )
        end 
        self:send(317, target, " ", idletime, " :seconds idle")
      end

      local client = nicks[lower(target)]
      if client then
        any_match = true
        match = true
        local idletime = math.floor(0.5 + moonbridge_io.timeref() - client.activity)
        self:send(311, client.nick, " ", client.user, " ", client.peer, " * :", client.real)
        local line
        for channel_name in pairs(client.channels) do
          if line then
            local line_new = line .. " " .. channel_name
            if #line_new > 56 then
              self:send(319, client.nick, " :", line)
              line = channel_name
            else
              line = line_new
            end
          else
            line = channel_name
          end
        end
        if line then
          self:send(319, client.nick, " :", line)
        end
        self:send(312, client.nick, " ", server_name, " :", server_version)
        self:send(317, client.nick, " ", idletime, " :seconds idle")
      end
      if not match then
        no_match[#no_match+1] = target
      end
    end
    if any_match then
      self:send(318, targets, " :End of WHOIS list")
    end
    for idx, target in ipairs(no_match) do
      self:send(401, target, " :No such nick/channel")
    end
    return true

  -- handle messages from irc client
  elseif self:is_command("PRIVMSG", 2) then

    local nick = upper(self.nick)

    -- check if nick is valid callsign
    if not string.match(nick, CALLSIGN_DETECT_REGEX) then
      return
    end

    -- extract callsign and ssid from nick
    local callsign, ssid = extract_callsign(nick)
    if not callsign then
      return
    end
    local station = normalize_station(callsign .. "-" .. ssid)
    local ax25_frame = prepare_ax25_frame(station)

    local aprs_data

    -- handle # channels
    if string.sub(self.params[1], 1, 1) == "#" then
      local channel_name = lower(string.sub(self.params[1], 2, #(self.params[1])))
      if #channel_name > max_channel_length then
        return
      end
      local message = self.params[2]
      send_chat(station, channel_name, message)

    -- handle + channels
    elseif string.sub(self.params[1], 1, 1) == "+" then
      local channel_name = upper(string.sub(self.params[1], 2, #(self.params[1])))
      if #channel_name > 9 then
        return
      end
      local message = self.params[2]
      if #message > 67 then
        message = string.sub(message, 1, 67)
      end
      local id = 1
      aprs_data = aprs.compose_bulletin(channel_name, message, id)

    -- handle & channels
    elseif string.sub(self.params[1], 1, 1) == "&" then
      local channel_name = lower(string.sub(self.params[1], 2, #(self.params[1])))
      if channel_name == "raw-tx" then
        aprs_data = self.params[2]
      end

    -- handle direct messages
    else
      local to = upper(self.params[1])
      local message = self.params[2]
      if #to > 9 then return end
      if #message > 67 then
        message = string.sub(message, 1, 67)
      end
      queue_message(self, station, to, message)

    end

    -- send out frame, if any
    if aprs_data then
      ax25_frame.info = aprs_data
      local ax25_data = ax25.compose_frame(ax25_frame)
      local kiss_data = kiss.compose_frame(ax25_data)
      io.write("tx ")
      ax25.debug(ax25_frame)
      local frame = ax25.parse_frame(ax25_data)
      table.insert(frames_heared, {
        frame = frame.source.station .. " " .. frame.destination.station .. " " .. frame.info,
        received = os.time()
      })
      send_kiss_frame(kiss_data)
    end

    if channels[lower(self.params[1])] then
      return false
    end
    return true
  end

end


